/*
 *   Burst error generation program (Gilbert model version)
 *   Copyright (C) 1996 NTT Mobile Communications Network, Inc. (DoCoMo)
 *   All right reserved.
 *                      Toshio  Miki     (miki@mlab.nttdocomo.co.jp)
 *                      Toshiro Kawahara (kawahara@mlab.nttdocomo.co.jp)
 *  Name:  gilbert
 *  
 *  Usage: gilbert <iseed> <bitrate(bps)> <average ber> <burst len(ms)> <bad_ber> <ifile> <ofile>
 * 
 *  Example usage:
 *      - Generate 10000 byte length error pattern sequence
 *         BER in burst duration: 50%
 *         Average bit error rate: 1.0e-3
 *         Bitrate = 24k bits/sec
 *         Burst length = 10 ms
 *             % gilbert 10 24000 1.0e-3 10 0.5 input.seq output.seq
 */
#include <stdio.h>
#define EFREESEC 0

#define GOOD_BER 0.0
/* #define BAD_BER  0.1*/

int putbit();
static int nerror = 0;
static int outbyte = 0;
int len;

void
main(argc, argv)
     int argc;
     char **argv;
{
  float frand();
  long iseed;
  double bitrate;
  int state;
  int flen, elen;
  int sum_flen, sum_elen, n_e, n_f;
  double freelen, burstlen;
  double badber;
  double aveber;
  int i;

  double p0, p1;
  double e0, e1;
  FILE *fpin, *fpout;
  char pat;

  if(argc != 8){
    fprintf(stderr, "Usage: %s <iseed> <bitrate(bps)> <average ber> <burst len(ms)> <bad_ber> <ifile> <ofile>\n", argv[0]);
    exit(1);
  }

  sscanf(argv[1], "%ld", &iseed);
  sscanf(argv[2], "%lf", &bitrate);
  sscanf(argv[3], "%lf", &aveber);
  sscanf(argv[4], "%lf", &burstlen);
  sscanf(argv[5], "%lf", &badber);

  if(argv[6][0] == '-')
    fpin = stdin;
  else if(argv[6][0] == '+'){
    fpin = NULL;
    sscanf(&argv[6][1], "%d", &len);
  }else{
    if(NULL == (fpin = fopen(argv[6],"rb"))){
      perror(argv[6]);
      exit(-1);
    }
  }

  if(argv[7][0] == '-')
    fpout = stdout;
  else{
    if(NULL == (fpout = fopen(argv[7],"wb"))){
      perror(argv[7]);
      exit(-1);
    }
  }
    
  burstlen = burstlen * bitrate / 1000.0;
  p1 = 1.0 / burstlen;
  p0 = aveber * p1 / (badber - aveber);
  freelen = 1.0 / p0;

#if 0  
  fprintf(stderr, "len = %d\n", len);
  fprintf(stderr, "average BER = %e\n", aveber);
  fprintf(stderr, "freelen = %f\n", freelen);
#endif

  flen = elen = sum_elen = sum_flen = n_e = n_f = 0;
  state = 0;

  for(i=0; i<bitrate*EFREESEC; i++)
    if(putbit(fpin, fpout, 0)){
      fprintf(stderr, "Finished while beginning error free period!!\n");
      exit(1);
    }

  for(;;){
    if(state == 0){		/* GOOD */
      if(putbit(fpin, fpout, (frand(&iseed) < GOOD_BER)))
	break;
      if(frand(&iseed) < p0){	/* TRANSITION */
	state = 1;
/*	printf("flen = %d\n", flen);*/
	sum_flen += flen;
	n_f ++;
	flen = 0;
      }else{
	flen ++;
      }
    }else{			/* BAD */
      if(putbit(fpin, fpout, (frand(&iseed) < badber)))
	break;
      if(frand(&iseed) < p1){	/* TRANSITION */
	state = 0;
/*	printf("elen = %d\n", elen);*/
	sum_elen += elen;
	n_e ++;
	elen = 0;
      }else{
	elen ++;
      }
    }
  }
  fprintf(stderr, "Statistics of generated error pattern:\n");
  fprintf(stderr, "\tAverage error free length:\t %f ms\n", (double)sum_flen/n_f/bitrate*1000.0);
  fprintf(stderr, "\tAverage burst length:\t\t %f ms\n", (double)sum_elen/n_e/bitrate*1000.0);
  fprintf(stderr, "\tAverage BER:\t\t\t %e\n", (double)nerror/(outbyte*8-EFREESEC*bitrate));
}

int
putbit(fpin, fpout, bpat)
     FILE *fpin, *fpout;
     int bpat;
{
  static int pos = 0;
  static char outdata = 0;
  static char indata;

  if(fpin != NULL && pos == 0){
    if(fread(&indata, sizeof(char), 1, fpin) != 1) /* EOF */
      return(1);		/* END!! */
  }else if(fpin == NULL && len == 0)
    return(1);			/* END!! */
  
  nerror += bpat;
  outdata <<= 1;
  outdata |= (bpat& 1);
  pos ++;
  if(pos == 8){			/* Write Out!! */
    if(fpin != NULL) outdata ^= indata;
    fwrite(&outdata, sizeof(char), 1, fpout);
    len --;
    outbyte ++;
    pos = 0;
  }
  return(0);
}
